"""
Global Filters
``````````````

The following are functions provided to Fyoo's jinja
runtime as jinja filters.
"""
import json

import yaml

__all__ = [
    'toJson',
    'toYaml',
]

toJson = json.dumps
"""Alias of json.dumps

.. code-block:: console

   $ fyoo -- echo "{{ [{'a': 1}, {'b': 2}] | toJson }}"
   [{"a": 1}, {"b": 2}]

"""
toYaml = yaml.safe_dump
"""Alias of yaml.safe_dump

.. code-block:: console

   $ fyoo -- echo "{{ [{'a': 1}, {'b': 2}] | toYaml }}"
   - a: 1
   - b: 2

"""





class FyooBaseException(Exception):
    """Base exception for fyoo library

    This exception is not to be used directly,
    but catching this error would catch a known exception raised
    by this internal library.
    """


class FyooTemplateException(FyooBaseException):
    """Exception raised when rendering templates
    """



from argparse import ArgumentParser
import json
from typing import List, Optional, Tuple, Union
import os

import jinja2
import yaml

from .template import (
    filters as fyoo_filters,
    attributes as fyoo_attributes,
)

_FYOO_SET_PREFIX = 'FYOO__SET__'


def implicit_type(string: str) -> Union[bool, int, float, str]:
    if string.lower() in {'true', 'false'}:
        if string.lower() == 'true':
            return True
        return False
    try:
        return int(string)
    except ValueError:
        pass
    try:
        return float(string)
    except ValueError:
        pass
    return string


def _parse_template_context(context_format: str, context_string: str):
    parse_methods = {
        'json': json.loads,
        'yaml': yaml.safe_load,
    }
    if context_format in parse_methods:
        result = parse_methods[context_format](context_string)
        if isinstance(result, dict):
            return result
        raise ValueError(f"Context was not a dictionary with format '{context_format}'")
    raise ValueError(f"Unrecognized context format '{context_format}'")


def _generate_fyoo_context(
        context_format: str,
        context_strings: Optional[List[str]],
        additional_vars: Optional[List[Tuple[str, str]]],
) -> dict:
    result_template = {}

    for key in filter(lambda k: k.startswith(_FYOO_SET_PREFIX) and len(k) > len(_FYOO_SET_PREFIX), os.environ.keys()):
        result_template[key[len(_FYOO_SET_PREFIX):]] = implicit_type(os.environ[key])

    if context_strings:
        for context_string in context_strings:
            result_template.update(_parse_template_context(context_format, context_string))
    if additional_vars:
        for key, value in additional_vars:
            result_template[key] = implicit_type(value)
    return result_template


def _set_type(string) -> Tuple[str, str]:
    eq_index = string.index('=')
    return string[:eq_index], string[eq_index + 1:]


class _FyooSecretParser(ArgumentParser):

    def __init__(self, *args, parent_parser: Optional = None, **kwargs) -> None:
        if not parent_parser:
            raise ValueError("'parent_parser' not provided")
        self.parent_parser = parent_parser
        super().__init__(*args, add_help=False, **kwargs)

    def add_argument(self, *args, **kwargs):
        self.parent_parser.add_argument(*args, **kwargs)
        super().add_argument(*args, **kwargs)


class FyooParser(ArgumentParser):

    """
    Just like an ArgumentParser, but adds a few hidden arguments. These
    hidden arguments will not show up in returned namespaces, rather they
    will simply tweak the full namespace before it comes back.
    """

    HELP = {
        'context': '''
Pass in a json or yaml string (multi-argument).
Can be set a single time by FYOO__CONTEXT.
'''.strip(),
        'context_format': '''
Context formatter to use.
Can be set by environment variable FYOO__CONTEXT_FORMAT.
'''.strip(),
        'set': r'''
Set a single context variable, i.e. table_name=users.

In addition, set context variables by environment variable
FYOO__SET__{var_name}={var_value}.

Context variables will try to use implicit types, defaulting
to string types.
        '''.strip(),
        'jinja_extension': '''Add a jinja2 extension to load at runtime.
        Can be set by environment variable FYOO__JINJA_EXTENSION, where ':' separates multiple extensions.''',
        'jinja_template_folder': '''Optionally, add a location for jinja2 to load templates from the filesystem.
        Can be set by environment variable FYOO__JINJA_TEMPLATE_FOLDER.''',
        'jinja_block_string': '''Jinja block start and end strings for blocks. Can be set
        by environment variables FYOO__JINJA_BLOCK_STRING__S/E''',
        'jinja_variable_string': '''Jinja block start and end strings for variables. Can be set
        by environment variables FYOO__JINJA_VARIABLE_STRING__S/E''',
        'jinja_comment_string': '''Jinja block start and end strings for comments. Can be set
        by environment variables FYOO__JINJA_COMMENT_STRING__S/E''',
    }

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        _F = FyooParser  # Short variable name for local access
        self.fyoo_secret_parser = _FyooSecretParser(parent_parser=self)
        self.fyoo_secret_parser.add_argument(
            '-c', '--context', action='append', help=_F.HELP['context'],
            default=[os.getenv('FYOO__CONTEXT')]
            if os.getenv('FYOO__CONTEXT')
            else [])
        self.fyoo_secret_parser.add_argument(
            '-f', '--context-format', help=_F.HELP['context_format'], default=os.getenv('FYOO__CONTEXT_FORMAT', 'json'))
        self.fyoo_secret_parser.add_argument(
            '-s', '--set', action='append', type=_set_type, help=_F.HELP['set'])
        self.fyoo_secret_parser.add_argument(
            '-je', '--jinja-extension', action='append', help=_F.HELP['jinja_extension'],
            default=[
                ext for ext in os.getenv('FYOO__JINJA_EXTENSION', '').split(':')
                if ext])
        self.fyoo_secret_parser.add_argument(
            '-jtf', '--jinja-template-folder', help=_F.HELP['jinja_template_folder'],
            default=os.getenv('FYOO__JINJA_TEMPLATE_FOLDER'))
        self.fyoo_secret_parser.add_argument(
            '-jbs', '--jinja-block-string', nargs=2, help=_F.HELP['jinja_block_string'],
            default=[os.getenv('FYOO__JINJA_BLOCK_STRING__S', '{%'),
                     os.getenv('FYOO__JINJA_BLOCK_STRING__E', '%}')])
        self.fyoo_secret_parser.add_argument(
            '-jvs', '--jinja-variable-string', nargs=2, help=_F.HELP['jinja_variable_string'],
            default=[os.getenv('FYOO__JINJA_VARIABLE_STRING__S', r'{{'),
                     os.getenv('FYOO__JINJA_VARIABLE_STRING__E', r'}}')])
        self.fyoo_secret_parser.add_argument(
            '-jcs', '--jinja-comment-string', nargs=2, help=_F.HELP['jinja_comment_string'],
            default=[os.getenv('FYOO__JINJA_COMMENT_STRING__S', '{#'),
                     os.getenv('FYOO__JINJA_COMMENT_STRING__E', '#}')])


    def parse_known_args(self, args=None, namespace=None):
        secret_known_args, secret_unknown_args = \
            self.fyoo_secret_parser.parse_known_args(args=args, namespace=namespace)

        loader = None \
                if secret_known_args.jinja_template_folder is None \
                else jinja2.FileSystemLoader(searchpath=secret_known_args.jinja_template_folder)
        jinja_env = jinja2.Environment(
            block_start_string=secret_known_args.jinja_block_string[0],
            block_end_string=secret_known_args.jinja_block_string[1],
            variable_start_string=secret_known_args.jinja_variable_string[0],
            variable_end_string=secret_known_args.jinja_variable_string[1],
            comment_start_string=secret_known_args.jinja_comment_string[0],
            comment_end_string=secret_known_args.jinja_comment_string[1],
            loader=loader,
        )

        jinja_env.globals.update({
            attr_name: getattr(fyoo_attributes, attr_name)
            for attr_name in fyoo_attributes.__all__
        })
        jinja_env.filters.update({
            attr_name: getattr(fyoo_filters, attr_name)
            for attr_name in fyoo_filters.__all__
        })

        if secret_known_args.jinja_extension:
            for fyoo_jinja_extension in secret_known_args.jinja_extension:
                jinja_env.add_extension(fyoo_jinja_extension)

        # Remove actions from current parser, as they're passed to fyoo inner secret-parser
        # pylint: disable=protected-access
        secret_action_dests = {action.dest for action in self.fyoo_secret_parser._actions}
        for action in list(self._actions):  # Need list to remove unstable iterable
            if action.dest in secret_action_dests:
                self._remove_action(action)

        template_context = _generate_fyoo_context(
            secret_known_args.context_format,
            secret_known_args.context,
            secret_known_args.set,
        )
        jinja_env.globals.update(template_context)

        known_args, unknown_args = super().parse_known_args(
            # Only use args that secret parser did not parse
            args=[
                jinja_env.from_string(arg).render()
                for arg in secret_unknown_args
            ],
            # But use initial optional namespace object
            namespace=namespace,
        )

        return known_args, unknown_args



import codecs
import itertools
import os
import re
import json

from setuptools import find_packages, setup
from setuptools.command.install import install

HERE = os.path.abspath(os.path.dirname(__file__))


def read(filename):
    with open(filename, 'r') as f:
        return f.read()


def read_code(*parts):
    with codecs.open(os.path.join(HERE, *parts), 'r') as fp:
        return fp.read()


def find_version():
    version_file = read_code('fyoo', '__init__.py')
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def install_requires():
    return [
        'jinja2',
        'PyYAML',
        'pytz',
    ]


def extras_require():
    require = {
        'test': [
            'pylint',
            'pytest',
            'pytest-cov',
            'pytest-env',
        ],
    }
    dev_requires = {
        'autopep8',
        'codecov',
        'coverage',
        'twine',
        'sphinx>=2.1',
        'sphinx-autobuild',
        'sphinx-argparse',
    }
    dev_requires.update(set(require['test']))  # dev tooling always needs test tooling
    require = dict(**require, dev=list(dev_requires))  # add in 'dev' extra
    # Ensure that all always has everything
    all_requires = list(set(
        itertools.chain.from_iterable(require.values())
    ))
    return dict(**require, all=all_requires)


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')
        version = find_version()

        if tag != version:
            raise RuntimeError(f'Git tag: {tag} does not match the version of this app: {version}')



class ShowInstallExtrasCommand(install):
    """Custom command to output extras"""
    description = 'show requirements and extras'

    def run(self):
        print(
f'''
Install requires
{json.dumps(install_requires(), indent=2)}

Extra requires
{json.dumps(extras_require(), indent=2)}
'''
        )


def do_setup():
    setup(
        author='Brian Kleszyk',
        name='fyoo',
        version=find_version(),
        description='A simple jinja2-based argument templatizer',
        long_description=read('README.rst'),
        long_description_content_type='text/x-rst',
        url='https://github.com/brian-bk/fyoo',
        project_urls={
            'Documentation': 'https://fyoo.readthedocs.io',
            'Source': 'https://github.com/brian-bk/fyoo',
            'Tracker': 'https://github.com/brian-bk/fyoo/issues',
        },
        packages=find_packages(exclude=['tests*']),
        install_requires=install_requires(),
        extras_require=extras_require(),
        entry_points={
            'console_scripts': [
                'fyoo = fyoo.cli:main',
            ],
        },
        python_requires='>=3.7',
        cmdclass={
            'verify': VerifyVersionCommand,
            'show_install': ShowInstallExtrasCommand,
        },
        license='MIT',
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
        ],
    )


if __name__ == '__main__':
    do_setup()
